#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAX_TOKEN_SIZE	1024

const char *get_token(void);
const char *convert_token(const char *token);
int is_class(const char *token);
int is_libfunc(const char *token);

int look_ahead = -1;
int last_nonalpha = -1;
int verbose = 0;

struct class {
	char *name;
	struct class *next;
} *clist;

int main(int argc, char **argv) {
	FILE *fp;
	const char *classfilename = 0;
	const char *token;
	int i;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'c':
				classfilename = argv[++i];
				break;

			case 'v':
				verbose = 1;
			default:
				break;
			}
		}
	}

	clist = calloc(1, sizeof(struct class));

	if(classfilename && (fp = fopen(classfilename, "r"))) {
		char name[256];
		while(fgets(name, 256, fp)) {
			struct class *cnode;
			int len = strlen(name);
			if(name[len - 1] == '\n') name[--len] = 0;

			cnode = malloc(sizeof(struct class));
			cnode->name = malloc(len + 1);
			strcpy(cnode->name, name);
			cnode->next = clist->next;
			clist->next = cnode;
			
		}
	}

	if(verbose) {
		struct class *cptr = clist->next;
		while(cptr) {
			fprintf(stderr, "class: %s\n", cptr->name);
			cptr = cptr->next;
		}
	}

	while((token = get_token())) {
		if(look_ahead == '(') {
			fputs(convert_token(token), stdout);
		} else {
			fputs(token, stdout);
		}
	}

	/* totally pointless code ... */
	while(clist) {
		struct class *tmp = clist;
		clist = clist->next;
		free(tmp->name);
		free(tmp);
	}

	return 0;
}


const char *get_token(void) {
	static char token[MAX_TOKEN_SIZE];
	int c;

	look_ahead = -1;
	
	while((c = getchar()) != EOF) {
		if(!isalpha(c)) {
			putchar(c);
			last_nonalpha = c;
		} else {
			int i = 0;
			do {
				token[i++] = c;
				if(i == MAX_TOKEN_SIZE) {
					fprintf(stderr, "Huge tokens!, output might be corrupted, aborting.\n");
					exit(-1);
				}
			} while((c = getchar()) != EOF && (isalnum(c) || c == '_'));
			
			token[i] = 0;
			look_ahead = c;
			ungetc(c, stdin);

			return token;
		}
	}
	return 0;
}


const char *convert_token(const char *intoken) {
	static char token[MAX_TOKEN_SIZE];
	char *out = token;
	const char *in = intoken;

	while(*in) {
		if(isupper(*in) && !is_class(intoken) && !is_libfunc(intoken)) {
			if(out > token && *(out - 1) != '_' && in > intoken && islower(*(in - 1))) {
				*out++ = '_';
			}
			*out++ = tolower(*in++);
		} else {
			*out++ = *in++;
		}
	}

	*out = 0;
	return token;
}

int is_class(const char *token) {
	struct class *cptr;
	
	cptr = clist->next;
	while(cptr) {
		if(!strcmp(token, cptr->name)) {
			return 1;
		}
		cptr = cptr->next;
	}

	return 0;
}

int is_libfunc(const char *token) {
	/* check for OpenGL functions (this will also catch GLX functions) */
	if(token[0] == 'g' && token[1] == 'l' && isupper(token[2])) {
		return 1;
	}

	/* check for SDL functions */
	if(token[0] == 'S' && token[1] == 'D' && token[2] == 'L' && token[3] == '_') {
		return 1;
	}

	/* check for Xlib functions */
	if(token[0] == 'X' && isupper(token[1])) {
		return 1;
	}

	/* check for FreeType functions */
	if(token[0] == 'F' && token[1] == 'T' && token[2] == '_') {
		return 1;
	}

	return 0;
}
